
/****************************************************************************/
//                           ***   ZXTAPE.C   ***
/*
	      TRANSFER ZX81 ".P" FILES FROM PC TO ZX81 OR TAPE
	     AND TRANSFER ZX81 ".P" FILES FROM A TAPE TO THE PC
     Original hardware & software concepts and original ZXTAPE by W.Rigter
	  Source code ported to 'C' from BASIC and versions 2.X, 3.X
  Copyright (C) Dec 1998 by Roger DiPaolo   email: dipaolor@eudoramail.com
	       Visit the 'ZX81 Virtual Community' web site at:
	     http://www.geocities.com/SiliconValley/Ridge/9965/

*/

//    Ported from BASIC to C by Roger DiPaolo 12/98 (dipaolor@eudoramail.com)
//    C compiler used is Borland C/C++ 3.1

//    The following LPT signals are used:
//    LPT data register (ioport0) - write data to data port (D0-D7).
//    Pin 2 - DATA BIT 0

//  version 2.0	12/12/98 rld	Initial release of 'C' version

//  version 2.1	12/19/98 rld	Fixed file listing to use all of screen
//				instead of scrolling one file name per
//				line.

//				Added loading of actual filename to ZX81.

//				Calibration is no longer run before each
//				load, to save time.  It is only run at the
//				start of the program.  There should be no
//				reason that timing should change after the
//				initial calibration.

//  version 2.2	12/20/98 rld	Added ability to select any I/O port address

//  version 2.3 12/24/98 rld    Added automatic parallel port detection

//  version 3.0 12/26/98 rld    Added save from tape to PC feature

//  version 3.1 01/02/99 rld    Added conditional compilation to provide
//                              for both positive and negative logic inputs
//                              on parallel port input pin 12.


/****************************************************************************/


// The following definition is the "normal", used with the standard ZXTAPE
// cable described in the 'zxtape.txt' document.  It is to be commented out
// when using an interface cable that inverts the input bit polarity (uses
// negative logic).  Also note that the parallel port's input at pin 12 is
// inverted from it's true "TTL" electrical state, therefore the code
// for "Positive Logic" is actually looking for a "1 level" as a "0".

#define POSITIVE_INPUT_LOGIC 1

#include "stdio.h"
#include "conio.h"
#include "dos.h"
#include "string.h"
#include "dir.h"
#include "process.h"
#include "ctype.h"
#include "time.h"
#include "alloc.h"
#include "io.h"
#include "math.h"

// Definitions

#define SUCCESS 1
#define FAILURE 0
#define YES 1           
#define NO 0
#define ON 1
#define OFF 0
#define OK 1
#define NOT_OK 0
#define QUIT 0
#define CONTINUE 1

#define ESCAPE 0x1B     // Escape
#define CTRL_LEFTBRACKET ESCAPE
#define CTRL_E 5		// Enquiry
#define ENQUIRY CTRL_E
#define CTRL_X 0x18		// Cancel
#define CANCEL CTRL_X
#define BELL 7          // ring bell through tty print
#define ZERONULL 0
#define BACKSPACE 8		// Backspace
#define SPACE 0x20		// Space
#define CARRIAGE_RETURN 0x0d

#define ECHO_ON 1
#define ECHO_OFF 0

#define BAD_CHOICE 0

#define BAD_DATA NOT_OK

#define MAXLINELEN 80
#define MAXDIRLEN 255

#define FUDGE_FACTOR 1.68  // Maybe I should be ashamed, but it works ;)

enum BOOL_ENUM {TRUE = 1, FALSE = 0};

typedef enum BOOL_ENUM BOOL;

// Prototypes

void getkey(void);
void dir1(void);
void getfile(void);
unsigned long calibrate_time(void);
void do_calibration(void);
unsigned GetHexValue(char *prompt_string, unsigned *status, BOOL NewLine);
int GetStrN(char *buffer,int echo,int maxchars);
void SetIOAddress(void);
void InitIOAddress(void);
void savefile(void);
void pause(char *msg);


// Globals

unsigned char 		*a = 0;
char			filestring[MAXLINELEN+1];
char			dirstring[MAXDIRLEN+1];
char			dir1string[MAXDIRLEN+1];
char			program_name[MAXLINELEN+1];
char			address_string[MAXLINELEN+1];
char			bytestring[MAXLINELEN+1];
char			inchar;
char			*dirptr;
int			ioport0, bytes, p;
unsigned int		n, timeout, looptime;
unsigned char		byte;
FILE			*infile;
int			tempword;
unsigned long		time_loops_per_ms;
unsigned long		pulse_hi_delay, pulse_low_delay, bit_delay;
float			ff = FUDGE_FACTOR;
struct text_info	ti;
int			tempchar;

/****************************************************************************/

main()
{
	unsigned long	i;
	unsigned char	maskbit = 1, b;
	unsigned int	l;
	char		dummy;


	clrscr();

#ifdef POSITIVE_INPUT_LOGIC

	printf("\n\n                                ZXTAPE v3.1p\n");

#else

	printf("\n\n                                ZXTAPE v3.1n\n");

#endif

	printf("\n              TRANSFER ZX81 \".P\" FILES FROM PC TO ZX81 OR TAPE\n");
	printf("\n             AND TRANSFER ZX81 \".P\" FILES FROM A TAPE TO THE PC\n");
	printf("\n\n     Original hardware & software concepts and original ZXTAPE by W.Rigter");
	printf("\n           Source code ported to 'C' from BASIC and versions 2.X, 3.X");
	printf("\n   Copyright (C) Dec 1998 by Roger DiPaolo   email: dipaolor@eudoramail.com\n");
	printf("\n\n               Visit the 'ZX81 Virtual Community' web site at: ");
	printf("\n             http://www.geocities.com/SiliconValley/Ridge/9965/");


	getkey();

	clrscr();

// set default values

	strcpy(filestring,"p.p");
	ioport0 = 0x378;
	InitIOAddress();
	dirptr = getcwd(dirstring,MAXLINELEN);

// Perform timing loop calibration

	do_calibration();

mainloop:

// initialize

	outp(ioport0,0);
	gettextinfo(&ti);

//    print menu

cont:

	clrscr();

	printf("\n Select one of the following ");

	printf("\n");
	printf("\n L) LOAD       ( Load a file to the ZX81 )");
	printf("\n S) SAVE       ( Save a file from tape to PC )");
	printf("\n F) FILE LIST  ( List P files in current directory )");
	printf("\n C) CHANGE DIR ( Current directory is %s )",dirstring);
	printf("\n B) BATCH      ( Call ZXTAPE.BAT )");
	printf("\n D) DOS        ( Type EXIT to return to ZXTAPE)");
	printf("\n I) IOPORT     ( Setup IOPORT : Current I/O port addr = %X)",ioport0);
	printf("\n H) HELP       ( About ZXTAPE )");
	printf("\n Q) QUIT       ( Return to DOS )");
	printf("\n");

	getkey();

	if(inchar == 'l')
	{
		goto load;
	}

	if(inchar == 's')
	{
		savefile();
	}

	if(inchar == 'f')
	{
		dir1();
		getkey();
	}

	if(inchar == 'b')
	{
		execlp("command.com","zxtape.bat",NULL);
		chdir(dirstring);
	}

	if(inchar == 'd')
	{
		printf("\nType EXIT to return from DOS");
		execlp("command.com",NULL);
	}

	if(inchar == 'i')
	{
		goto ioport;
	}

	if(inchar == 'c')
	{
		printf("\nEnter new working directory: ");
		gets(dir1string);
		if(!dir1string[0])
		{
			goto end;
		}
		else
		{
			strncpy(dirstring,dir1string,MAXDIRLEN);
		}
		chdir(dirstring);
	}

	if(inchar == 'h')
	{
		clrscr();

//		printf("\nHARDWARE:");
//		printf("\nTurn your PC into a cassette player loading ZX81 programs with ZXTAPE");
//		printf("\nJust make up a simple cable with 2 wires and a male DB25 connector");
//		printf("\n(for the PC printer port) to one end and a 3.5mm headphone-style plug");
//		printf("\n(for the ZX81 EAR jack) on the other end. Connect one wire to DB25");
//		printf("\npin 2 and TIP terminal of a 3.5mm headphone-style plug. Connect the");
//		printf("\nother wire to DB25 pin 24 and the RING (or ground) terminal of the\nplug.  That's it!");

//		printf("\nSOFTWARE:");

		printf("\nThe serial data generated by ZXTAPE simulates the normal tape signal");
		printf("\nrecorded by a ZX81. The pulses are generated with delays based on ");
		printf("\nsoftware loops which are calibrated at program initialization.  This should");
		printf("\ninsure accurate timing regardless of the speed of your PC's CPU.");

		printf("\n\nFor instructions on how to make a simple 'ZXTAPE' cable that connects");
		printf("\nyour PC's parallel port to the ZX81 and tape recorder, see the accompanying");
		printf("\nfile 'ZXTAPE.TXT'.  If you do not have that text file, you can download it");
		printf("\nalong with the latest version of ZXTAPE from the 'ZX81 Virtual Community'");
                printf("\nweb site at: http://www.geocities.com/SiliconValley/Ridge/9965/");

		printf("\n***** COMMAND SUMMARY *****");
		printf("\nL - On the PC type in a file name that you wish to load from the P file\n    listing, press enter, then follow instructions.");
		printf("\nF - Lists all the .P files in the current directory");
		printf("\nC - enter new directory : for example \\zx81\\files");
		printf("\nB - run batch file ZXTAPE.BAT in the current directory:");
		printf("\n    for example create a simple ASCII file with the text");
		printf("\n");
		printf("\n    XTenders");
		printf("\n");
		printf("\n    to run XTender from within the ZXTAPE program.");
		printf("\nD - Shell out to DOS and return by typing EXIT");
		printf("\nI - Select IOPORT address for your parallel port");

		printf("\n--> HIT ANY KEY TO RETURN TO MENU...");

		while(!kbhit());
	}

	if(inchar == 'q')
	{
		clrscr();
		exit(1);
	}

	goto mainloop;


load:
	getfile();
	if(bytes == 0)
	{
		printf("\nError attempting to load file");
		delay(3000);
		clrscr();
		goto mainloop;
	}


	clrscr();

outfile:

	printf("\nOn your ZX81 enter command: LOAD \"%s\" ,then hit any key on PC",program_name);
	pause("");
	printf("\r                                                                                 ");
	printf("\rSending 5 seconds of silence...");
	outp(ioport0,0);
	delay(5000);  // lead in with 5 seconds of silence
	printf("\rTransfer will take about %d seconds",(int)ceil((18.0 * ((float)bytes)) / 1000.0));
	printf("\n*** If you wish to quit transfer, press space bar to abort ***");
	printf("\nLoading %d bytes of file %s",bytes,filestring);

	// Remove any "stray" keyhit from the key buffer that may cause
	// 'kbhit()' to take more processing time than usual, therefore
	// slowing down the main loop which would result in incorrect
	// byte xfer timing.

	if(kbhit())
	{
		tempchar = getch();
	}

	// This is the main xfer loop

	for(n=0;n<bytes;n++)  // Once through this loop for each byte
	{

		if(kbhit())  // Check for user abort here
		{
			if(getch() == ' ')
			{
				goto breakout;
			}
		}

		maskbit = 0x80;  // Start each byte by sending msb

		do  // Once through this loop for each bit
		{
			if(!(a[n] & maskbit))  // Determine if bit is 0 or 1
			{
			   p=4;  // A zero is 4 pulses
			}
			else
			{
			   p=9;  // A one is 9 pulses
			}

			for(l=0;l<p;l++)  // Output either 4 or 9 pulses
			{
				outp(ioport0,1);
				for(i=0;i<pulse_hi_delay;i++);	// high time is about .15ms
				outp(ioport0,0);
				for(i=0;i<pulse_low_delay;i++);	// low time is about .15ms
			}

			// Delay between bits

			for(i=0;i<bit_delay;i++);  // delay about 1.5ms before next bit

			maskbit >>= 1;  // Do next least signifigant bit

		}while(maskbit > 0);
	}

outfile1:

	printf(" - loading finished! ");
	goto end;

breakout:

	printf(" - load interrupted!");
	goto end;

ioport:

	SetIOAddress();

end:

	if(a)
	{
		free(a);
		a = (unsigned char *)0;
	}
	fclose(infile);
	printf("\n ");
	printf("\nWait 5 seconds, or press any key to continue...");

	timeout = 0;

	while(!kbhit())
	{
		delay(10);

		if(timeout++ == 500)  // Wait 5 seconds for a keyhit before 
		{		      // going automatically back
			break;	      // to mainloop
		}

	}

	goto mainloop;

}

/****************************************************************************/

void
getkey()
{
	inchar = tolower(getche());
}

/****************************************************************************/

void
dir1()
{
	int 			done, line=2, column=1;
	int			dummy;
	struct find_t		ffblk;

	clrscr();
	printf("List of .P files in current directory :");
	done = _dos_findfirst("*.p",_A_NORMAL,&ffblk);

	while(!done)
	{
		gotoxy(column,line);
		printf("%s",ffblk.name);
		done = _dos_findnext(&ffblk);
		if(column <= 60)
		{
			column += 20;
		}
		else
		{
			line++;
			if(line == (ti.screenheight - 1))
			{
				gotoxy(1,ti.screenheight);
				printf("***** HIT ENTER KEY TO DISPLAY MORE FILES *****");
				dummy = getchar();
				clrscr();
				line = 1;
			}
			column = 1;
		}
	}
}

/****************************************************************************/

void
getfile()
{
	long	flen;
	int	program_name_len, i;


	clrscr();

	// Let user choose file

	dir1();
	printf("\nEnter name of data file without extension: ");
	gets(program_name);

	if(program_name != "")
	{
	   strncpy(filestring,program_name,MAXLINELEN);
	   strcat(filestring,".p");
	}
	else
	{
		return;
	}

	// Open file
	
	if(!(infile = fopen(filestring,"rb")))
	{
		sound(440);
		delay(500);
		nosound();
		printf("\nFile not found");
		return;
	}

	// Measure file length (in bytes)

	fseek(infile,0,SEEK_END);
	flen = ftell(infile);
	fseek(infile,0,SEEK_SET);

	program_name_len = strlen(program_name);

	// Allocate a buffer large enough to hold entire file

	a = (unsigned char *)malloc(flen+program_name_len+2);

	if(!a)
	{
		printf("\nNot enough memory to load file");
		return;
	}

	// Convert ASCII file name to "Sinclair Code" program name and store

	for(i=0;i<program_name_len;i++)
	{
		program_name[i] = toupper(program_name[i]);
		a[i] = program_name[i] - 27;
	}

	a[i-1] += 128;  // Make last character of program name "inverse"

	// Read file and load into buffer, 1 byte at a time

	for(bytes=i;;bytes++)
	{
		tempword = fgetc(infile);
		if(tempword == EOF)
		{
			break;
		}
		else
		{

			a[bytes] = (unsigned char)tempword;
		}
	}

}

/****************************************************************************/

unsigned long
calibrate_time()
{
	unsigned long	i, loops = 50000000;
	unsigned	start, end;
	unsigned long	IterationsPerMillisecond;

	printf("\nPerforming time calibration, please wait....");

	start = (unsigned)clock();
	for(i=0;i<loops;i++);
	end = (unsigned)clock();
	looptime = end - start;  // number of 55ms time chunks in loop
	IterationsPerMillisecond = (unsigned long)(loops / (unsigned long)(looptime*55));

	clrscr();

	return(IterationsPerMillisecond);

}

/****************************************************************************/

void
do_calibration(void)
{
	time_loops_per_ms = calibrate_time();

	printf("\n%d time loops per ms, each loop is %f ms",time_loops_per_ms,(float)(1.0/(float)time_loops_per_ms));

	bit_delay = (unsigned long)((float)time_loops_per_ms * 1.2 * ff);
	pulse_hi_delay = (unsigned long)((float)time_loops_per_ms * 0.13 * ff);
	pulse_low_delay = (unsigned long)((float)time_loops_per_ms * 0.11 * ff);
}

/*****************************************************************************
** NAME:	GetHexValue
**
** PURPOSE:		Retrieve numerical user input in hexadecimal format			
**
** PROTOTYPE:	unsigned GetHexValue(char *prompt_string, unsigned *status, BOOL NewLine)	
**
** ARGUMENTS:	
**		char		*prompt_string		-  pointer to string to prompt user
**		unsigned	*status				-  a pointer to an unsigned word  
**											  
**			The following data is stored in *status for processing by calling
**			function:
**					ESCAPE				- escape key was pressed.
**					BAD_DATA			- no data entered or data not in hex
**										format.
**					OK 					- none of the previously described 
**										events occurred.
**
**		BOOL		NewLine			- if TRUE insert a newline before the prompt.
**
** RETURN VALUE:	Returns hexadecimal value entered OR returns a 0
**						for the following conditions:
**						1.  Escape key was pressed.
**						2.  No characters or more than 8 characters entered when	
**							prompted for input.
**						3.  Data entered is not in hexadecimal format. 
**
** GLOBALS USED:	None	
**
** REMARKS:	
**
******************************************************************************/

unsigned
GetHexValue(char *prompt_string, unsigned *status, BOOL NewLine)
{
	char		buffer[11];
	int			num_of_chars;
	int			GetStrStatus;
	unsigned	value;

	*status = OK;

	if(NewLine == TRUE)
	{
		printf("\n%s",prompt_string);
	}
	else
	{
		printf("%s",prompt_string);
	}

	GetStrStatus = GetStrN(buffer,ECHO_ON,8);
	
	if(GetStrStatus == ESCAPE)
	{
		*status = ESCAPE;
		return(0);										
	}

	buffer[10] = '\0';

	num_of_chars = strlen(buffer);

	if((num_of_chars > 8) || (num_of_chars < 1))
	{
		*status = BAD_DATA;
		return(0);										
	}	

	if(!sscanf(buffer,"%lx",&value))
	{

		// Error: data entered not in hexadecimal format.
		// Note: data does not have to be preceded by hex value delimiter '0x'
		*status = BAD_DATA;							   
		return(0);										
	}

	return(value);
}


/****************************************************************************/

int
GetStrN(char *buffer,int echo,int maxchars)
{
	int		charcount = 0;

	while(1)
	{
		buffer[charcount] = getch();
		switch(buffer[charcount])
		{
			case ESCAPE:
							putchar('\n');
							buffer[0] = '\0';
							return(ESCAPE);

			case BACKSPACE:
							if(charcount > 0)
							{
								if(echo == ECHO_ON)
								{
									putchar(BACKSPACE);
									putchar(SPACE);
									putchar(BACKSPACE);
								}
								charcount--;
							}
							break;

			case CARRIAGE_RETURN:
							putchar(0x0d);
							putchar(0x0a);
							buffer[charcount] = '\0';
							return(0);

			default:
							if((buffer[charcount] >= SPACE) && (buffer[charcount] <= 0x7e)) 
							{
								if(charcount < maxchars)
								{
									if(echo == ECHO_ON)
									{
										putchar(buffer[charcount]);
									}
									charcount++;
								}
							}
		}
	}
}

/****************************************************************************/

void
SetIOAddress(void)
{
	int					newportaddr;
	unsigned			status;
	unsigned far		*ptraddr = (unsigned far *)0x00000408;
	unsigned 	 		address;
	unsigned 			portadr[4];
	int					a, onefound = 0;

	clrscr();

	for(a=0;a<4;a++)
	{
		portadr[a] = (unsigned)*ptraddr;
		if(portadr[a] != 0)
		{
			printf("\nLPT%d found at I/O addess %X hex",a+1,portadr[a]);
			if(onefound == 0)
			{
				newportaddr = ioport0 = portadr[a];
			}
			onefound = 1;
		}
		ptraddr++;
	}

	if(onefound)
	{

getlptnum:

		printf("\nEnter the LPT # of the port you wish to use, or 0 to specify another: ");
		getkey();
		if((inchar >= '1') && (inchar <= '4') && (portadr[((inchar - '0') - 1)] != 0))
		{
			ioport0 = portadr[((inchar - '0') - 1)];
			goto SetIOAddressDone;
		}
		else if(inchar != '0')
		{
			sound(440);
			delay(500);
			nosound();
			goto getlptnum;
		}
	}
	else
	{
		printf("\nNo LPT ports found by ZXTAPE");
	}

	if((newportaddr = GetHexValue("Enter I/O port address (in hex): ",&status,TRUE)) == 0)
	{
		if(status == ESCAPE)
		{
			printf("\n>> Enter Aborted <<");
		}
		else if(status == BAD_DATA)
		{
			printf("\n>> Invalid data, Address unchanged <<");
		}
	}
	else
	{
		ioport0 = newportaddr;
	}

SetIOAddressDone:

	printf("\nI/O port address is %X hex",ioport0);
}

/****************************************************************************/
//
// InitIOAddress() - Set the current parallel port to the 1st found installed
//
// **************************************************************************

void
InitIOAddress(void)
{
	unsigned far		*ptraddr = (unsigned far *)0x00000408;
	unsigned 		portadr[4];
	int			a;

	for(a=0;a<4;a++)
	{
		portadr[a] = (unsigned)*ptraddr;
		if(portadr[a] != 0)
		{
			printf("\nUsing LPT%d at I/O addess %X hex",a+1,portadr[a]);
			ioport0 = portadr[a];
			break;
		}
		ptraddr++;
	}
}

/****************************************************************************/
// savefile() - Capture a ZX81 program and write it to the PC as a '.p' file
//
//
/****************************************************************************/

void
savefile()
{
	unsigned	program_name_len;
	unsigned  	i;
	unsigned char	*inbuf, *dataptr, maskbit = 0x80, sample;
	int		in_port = ioport0 + 1;
	int		dummy;
	unsigned 	hicount = 0, pulseperiod;
	unsigned 	pulsecount = 1, glitchperiod, decisionpoint;
	unsigned 	bytecount = 0, programnamelen;
	unsigned 	x = 0,y = 100;
	char		filename[80];
	unsigned 	buffersize = (1024 * 16);
	unsigned 	h;
	unsigned long	lowcount = 0, low_end, l;


	clrscr();

	inbuf = malloc(buffersize);

	dataptr = inbuf;

	for(i=0;i<buffersize;i++)
	{
		inbuf[i] = 0;
	}

	pause("\nHit any key on PC to begin saving data");

	sample = inportb(in_port);

#ifdef POSITIVE_INPUT_LOGIC

	if(!(sample & 0x20))
	{
		pause("\nERROR: Input line is initially high (logic '1')\n       A short leader of 'silence' must preceed data.");
		free(inbuf);
		return;
	}

#else

	if(sample & 0x20)
	{
		pause("\nERROR: Input line is initially low (logic '1')\n       A short leader of 'silence' must preceed data.");
		free(inbuf);
		return;
	}

#endif

	printf("\nCapturing...");

#ifdef POSITIVE_INPUT_LOGIC

	// Wait for initial "silence" to end

	do
	{
		sample = inportb(in_port);

	}while(sample & 0x20);

	// Get a measure of the first "high" period

	while(!(sample & 0x20))
	{
		sample = inportb(in_port);
		hicount++;
		if(x>y)
		{
		}

	}

	// Get a measure of the first "low" period

	while(sample & 0x20)
	{
		sample = inportb(in_port);
		lowcount++;
		if(x>y)
		{
		}
	}

#else

	// Wait for initial "silence" to end

	do
	{
		sample = inportb(in_port);

	}while(!(sample & 0x20));

	// Get a measure of the first "high" period

	while(sample & 0x20)
	{
		sample = inportb(in_port);
		hicount++;
		if(x>y)
		{
		}
	}

	// Get a measure of the first "low" period

	while(!(sample & 0x20))
	{
		sample = inportb(in_port);
		lowcount++;
		if(x>y)
		{
		}

	}

#endif


	h = hicount;
	l = lowcount;

	// Calculate an average "pulse" period

	pulseperiod = (hicount + lowcount) / 2;

	// Calculate the value below which we consider
	// the reading a "glitch"

	glitchperiod = pulseperiod / 4;

	// Calculate the value at which we make a decision
	// If a "lowcount" is less than this, we consider it
	// part of a pulse.  If it is greater than this we consider
	// it a bit-boundry.

	decisionpoint = pulseperiod * 5;

	// Calculate the length of "silence" we use to determine the end

	low_end = pulseperiod * 500;

	// Capture the input data

	hicount = 1;
	lowcount = 0;

	while(1)
	{

#ifdef POSITIVE_INPUT_LOGIC

		while(!(sample & 0x20))
		{
glitch:
			sample = inportb(in_port);
			hicount++;
			if(x>y)
			{
			}
		}

		if(hicount <= glitchperiod)
		{
			hicount = 0;
			goto glitch;  // ignore glitches
		}

		// Wait for the low period between pulses to pass

		while(sample & 0x20)
		{
			sample = inportb(in_port);
			lowcount++;
			if(lowcount > low_end)
			{
				break;
			}
		}

#else

		while(sample & 0x20)
		{
glitch:
			sample = inportb(in_port);
			hicount++;
			if(x>y)
			{
			}
		}

		if(hicount <= glitchperiod)
		{
			hicount = 0;
			goto glitch;  // ignore glitches
		}

		// Wait for the low period between pulses to pass

		while(!(sample & 0x20))
		{
			sample = inportb(in_port);
			lowcount++;
			if(lowcount > low_end)
			{
				break;
			}
		}


#endif

		if(lowcount <= decisionpoint)
		{
			pulsecount++;
		}
		else
		{
			// We are at a bit boundry

			if(pulsecount > 5)
			{
				*dataptr |= maskbit;
			}

			if(maskbit == 1)
			{
				maskbit = 0x80;
				dataptr++;
				bytecount++;
				if(bytecount >= buffersize)
				{
					printf("\nERROR: Buffer Full!\n");
					break;
				}
			}
			else
			{
				maskbit >>= 1;
			}

			if(lowcount > low_end)
			{
				break;  // File transfer complete!
			}

			pulsecount = 0;

		}

		hicount = 0;
		lowcount = 0;
	}


	printf("DONE.");
	printf("\nProcessing...");

	// Create the file name for this file by converting
	// "Sinclair Code" program name to DOS ASCII filename

	dataptr = inbuf;

	i = 0;

	while(1)
	{
		if(!(dataptr[i] & 0x80))
		{
			filename[i] = dataptr[i] + 27;
		}
		else  // Last character of program name is "inverse"
		{
			filename[i] = (dataptr[i] & 0x7f) + 27;

			// All characters not allowed by DOS converted to underscore

			if((filename[i] < 'A') || (filename[i] > 'Z'))
			{
				if((filename[i] < '0') || (filename[i] > '9'))
				{
					filename[i] = '_';
				}
			}

			i++;
			break;
		}

		// All characters not allowed by DOS converted to underscore

		if((filename[i] < 'A') || (filename[i] > 'Z'))
		{
			if((filename[i] < '0') || (filename[i] > '9'))
			{
				filename[i] = '_';
			}
		}

		if(i<70)
		{
			i++;
		}
		else
		{
			break;
		}
	}

	printf("DONE.");

	programnamelen = i;

	// If necessary, truncate filename at 8 characters

	if(i>8)
	{
		i = 8;
	}

	// Add ".p" suffix to create file name

	strcpy((filename+i),".p");

	printf("\nSaving %u bytes to file '%s'...",bytecount-programnamelen,filename);

	if(!(infile = fopen(filename,"wb")))
	{
		sound(440);
		delay(500);
		nosound();
		printf("\nFile open error!");
		free(inbuf);
		return;
	}

	fwrite((void *)(inbuf+programnamelen),(size_t)1,(size_t)(bytecount-programnamelen),infile);

	printf("\nDone, closing file.");

	fclose(infile);

	free(inbuf);

	pause("\nHit any key to return to menu");

}


// *************************************************************************

void
pause(char *msg)
{
	int	dummy;


	printf("%s",msg);
	while(!kbhit());
	dummy = getch();
}

